#!/usr/bin/env python

__doc__ = """
Implements the penalized logistic regression.
"""

__all__ = ["plr","predict","cv_plr","plr_pred","nscore_predict"]

from numpy import *
from scipy.io.matlab import *

def plr(X,d,lm=None,criterion=0.05,chance=0.):
    """Perform logistic regression
    
    Input
    =====
    X           a patterns X features data matrix
    d           a vector of row labels for every pattern
    lm          penalty term
    criterion   the desired error when finished
    chance      chance level
    
    Output
    ======
    w           Weight vector such that f(w[:-1].T*X+w[-1])
                is the prediction
    err         size of the last step
    k           number of steps
    """
    # Structure of the data
    
    if not lm:
        lm = cv_plr(X,d)
    npatterns,nfeatures = X.shape

    # Make everything matrix
    X = matrix(concatenate((X,ones((npatterns,1),'d')),-1)).T
    w = matrix(zeros((nfeatures+1,1),'d'))
    d = matrix(d)

    # This is where the errors go
    dw = w.copy()
    old_dw = 100+w.copy()

    # Perpare iteration
    Lm = lm * identity(nfeatures+1)
    Lm[-1,-1] = 0
    k = 0
    err = 100
    while err>criterion:
        # How did it change last time?
        old_dw[:] = dw

        # Now compute prediction
        p = (1-chance)*f(w.T*X) + chance

        # Gradient
        g = X*(d-p).T
        g[:-1] -= lm*w[:-1]

        # Fisher information matrix
        H = matrix(X.A*(p.A*(1-p.A)))*X.T
        H += Lm

        # Modify weights
        dw = H.I*g
        w += dw

        # Prepare next step
        k += 1
        err = sum((dw-old_dw).A**2)
    return w,err,k

def predict(Y,w,chance=0,ptype = 'class'):
    """Perform prediction with weights from plr
    
    Input
    =====
    Y           a patterns X features data matrix
    w           a weight vector as returned for instance by plr
    chance      chance level for prediction

    Output
    ======
    p           probabilities for class '1' for each pattern
    """
    npatterns,nfeatures = Y.shape
    X = matrix(concatenate((Y,ones((npatterns,1),'d')),-1)).T
    if ptype == 'response':
            return f(w.T*X)*(1-chance) + chance
    return (w.T*X>0)*1

def f(y):
    return 1./(1.+exp(-y))

def cv_plr(x,y,lamb=array([1,0.1,0.01,1e-3,1e-4,1e-5,1e-6]),nfold=5):
   nseq = len(y)
   m = len(lamb)
   folds = zeros([nseq,2],dtype = int)
   folds[:,0] = random.permutation(arange(nseq))
   folds[:,1] = arange(nseq)%nfold
   errorarr = zeros([m, nfold])
   aucarr = zeros([m,nfold])
   for i in range(nfold):
       trainsets = folds[folds[:,1]!=i][:,0]
       omit = folds[folds[:,1] == i][:,0]
       for j in range(m):
           fit = plr(x[trainsets,:],y[trainsets].T,lm = lamb[j])
           pred = predict(x[omit,:],fit[0],ptype = 'class')
           errorarr[j,i] = mean(pred != y[omit])
           pred = predict(x[omit,:],fit[0],ptype = 'response')
           pred = asarray(pred).flatten()
           idx_nuc = (y[omit] ==1)
           idx_link = (y[omit] ==0)
           aucarr[j,i] = auc_val(pred[idx_nuc],pred[idx_link])
           #print lamb[j],i, aucarr[j,i],errorarr[j,i]
       print "CV fold " + str(i) +"\n"
   #error = mean(errorarr, 1)
   #return error, erroarr
   auc = mean(aucarr,1)
   print "Train auc value is "+ str(max(auc))+" with lambda = "+str(lamb[auc == max(auc)])
   return float(lamb[auc == max(auc)])
   
# auc ##############################

def auc_val(nuc_score,link_score):
    #nnuc = nuc_score.shape[1]
    #nlink = link_score.shape[1]
    nnuc = len(nuc_score)
    nlink = len(link_score)
    nseq = nnuc + nlink
    sensitivity = zeros([nseq, 1])
    specificity = zeros([nseq, 1])
    sort_score = argsort(hstack((nuc_score,link_score)))[::-1]
    for i in range(1,nseq):
        ntp = (sort_score[0:i]<=nnuc).sum()*1.0
        n_p = i
        sensitivity[i-1] = ntp/nnuc
        specificity[i-1] = (nlink - n_p + ntp)/nlink
    auc = sum(sensitivity[1:]*(specificity[0:-1] - specificity[1:]))
    return auc

# plr #############################
def plr_pred(x,y,xnew,output):
    lamb = cv_plr(x,y.T)
    fit = plr(x,y.T,lm=lamb)
    pred_class = predict(xnew,fit[0])
    pred_response = predict(xnew,fit[0],ptype = 'response')
    savemat(output+'beta', {'beta': fit[0]},oned_as='row')
    return pred_class,pred_response

def nscore_predict(model,xnew,output):
    pred_class = predict(xnew,model)
    pred_response = predict(xnew,model,ptype = 'response')
    return pred_class,pred_response

 
   


  
    

